package terminal;

import static terminal.ControlSequenceHandler.isEscape;
import static util.Characters.EOF;

import java.io.*;
import java.util.concurrent.*;

import com.googlecode.lanterna.TerminalPosition;
import com.googlecode.lanterna.input.*;
import com.googlecode.lanterna.terminal.IOSafeTerminal;

public class LispTerminal {

    public static final char END_OF_SEGMENT = 'x';

    private IOSafeTerminal terminal;
    private PipedOutputStream inputWriter;
    private PipedInputStream outputReader;
    private ExecutorService executorService;
    private ControlSequenceHandler controlSequenceHandler;
    private TerminalPosition origin;
    private String inputLine;
    private String outputSegment;
    private boolean isFinished;

    public LispTerminal(IOSafeTerminal terminal, PipedOutputStream inputWriter, PipedInputStream outputReader) {
        this.terminal = terminal;
        this.inputWriter = inputWriter;
        this.outputReader = outputReader;
        this.executorService = Executors.newFixedThreadPool(2);
        this.controlSequenceHandler = new ControlSequenceHandler();
        this.origin = terminal.getCursorPosition();
        this.inputLine = "";
        this.outputSegment = "";
        this.isFinished = false;
    }

    public void run() {
        executorService.execute(this::readInput);
        executorService.execute(this::writeOutput);
        executorService.shutdown();
    }

    public void readInput() {
        while (!isFinished)
            processNextKey();
    }

    private void processNextKey() {
        handleKey(getKeyStroke());
        takeNap();
    }

    private KeyStroke getKeyStroke() {
        KeyStroke keyStroke = null;

        try {
            keyStroke = terminal.pollInput();
        } catch (IllegalStateException e) {
            // TODO - Issue #299
            terminal.putCharacter('\n');
            System.exit(0);
        }

        return keyStroke;
    }

    private synchronized void handleKey(KeyStroke keyStroke) {
        if (keyStroke == null)
            return;

        doKey(keyStroke);
        terminal.flush();
    }

    private synchronized void doKey(KeyStroke keyStroke) {
        KeyType keyType = keyStroke.getKeyType();

        if (keyStroke.isCtrlDown())
            doControlKey(keyStroke);
        else if (keyType == KeyType.ArrowLeft)
            moveCursorLeft();
        else if (keyType == KeyType.ArrowRight)
            moveCursorRight();
        else if (keyType == KeyType.Enter)
            doEnter();
        else if (keyType == KeyType.Backspace)
            doBackspace();
        else if (keyType == KeyType.Delete)
            doDelete();
        else if (keyType == KeyType.Character)
            doCharacter(keyStroke);
    }

    private synchronized void doControlKey(KeyStroke keyStroke) {
        KeyType keyType = keyStroke.getKeyType();

        if (keyType == KeyType.Character)
            doControlCharacter(keyStroke);
    }

    private synchronized void doControlCharacter(KeyStroke keyStroke) {
        if (keyStroke.getCharacter() == 'd')
            doControlD();
    }

    private synchronized void doControlD() {
        doEnter();
        finish();
    }

    private synchronized void moveCursorLeft() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveLeft(cursorPosition))
            retractCursor(cursorPosition);
    }

    private synchronized boolean isPossibleToMoveLeft(TerminalPosition cursorPosition) {
        return distanceFromOrigin(cursorPosition) > 0;
    }

    private synchronized int distanceFromOrigin(TerminalPosition cursorPosition) {
        int columnDifference = cursorPosition.getColumn() - origin.getColumn();
        int rowDifference = cursorPosition.getRow() - origin.getRow();
        int totalColumns = terminal.getTerminalSize().getColumns();

        return columnDifference + (totalColumns * rowDifference);
    }

    private synchronized void retractCursor(TerminalPosition cursorPosition) {
        if (isAtStartOfRow(cursorPosition))
            terminal.setCursorPosition(terminal.getTerminalSize().getColumns(), cursorPosition.getRow() - 1);
        else
            terminal.setCursorPosition(cursorPosition.withRelativeColumn(-1));
    }

    private boolean isAtStartOfRow(TerminalPosition cursorPosition) {
        return cursorPosition.getColumn() == 0;
    }

    private synchronized void moveCursorRight() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveRight(cursorPosition))
            advanceCursor(cursorPosition);
    }

    private synchronized boolean isPossibleToMoveRight(TerminalPosition cursorPosition) {
        return distanceFromOrigin(cursorPosition) < inputLine.length();
    }

    private synchronized void advanceCursor(TerminalPosition cursorPosition) {
        if (isAtEndOfRow(cursorPosition))
            terminal.setCursorPosition(0, cursorPosition.getRow() + 1);
        else
            terminal.setCursorPosition(cursorPosition.withRelativeColumn(1));
    }

    private synchronized boolean isAtEndOfRow(TerminalPosition cursorPosition) {
        return cursorPosition.getColumn() == terminal.getTerminalSize().getColumns() - 1;
    }

    private synchronized void doEnter() {
        terminal.putCharacter('\n');
        inputLine += "\n";

        try {
            inputWriter.write(inputLine.getBytes());
            inputWriter.flush();
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        inputLine = "";
        origin = terminal.getCursorPosition();
    }

    private synchronized void doBackspace() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveLeft(cursorPosition)) {
            String remaining = inputLine.substring(distanceFromOrigin(cursorPosition), inputLine.length());
            inputLine = inputLine.substring(0, distanceFromOrigin(cursorPosition) - 1) + remaining;

            retractCursor(cursorPosition);

            for (char c : remaining.toCharArray())
                terminal.putCharacter(c);

            terminal.putCharacter(' ');

            retractCursor(cursorPosition);
        }
    }

    private synchronized void doDelete() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveRight(cursorPosition)) {
            String remaining = inputLine.substring(distanceFromOrigin(cursorPosition) + 1, inputLine.length());
            inputLine = inputLine.substring(0, distanceFromOrigin(cursorPosition)) + remaining;

            for (char c : remaining.toCharArray())
                terminal.putCharacter(c);

            terminal.putCharacter(' ');
            terminal.setCursorPosition(cursorPosition);
        }
    }

    private synchronized void doCharacter(KeyStroke keyStroke) {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveRight(cursorPosition)) {
            String remaining = keyStroke.getCharacter()
                               + inputLine.substring(distanceFromOrigin(cursorPosition), inputLine.length());
            inputLine = inputLine.substring(0, distanceFromOrigin(cursorPosition)) + remaining;

            for (char c : remaining.toCharArray())
                terminal.putCharacter(c);

            advanceCursor(cursorPosition);
        } else {
            terminal.putCharacter(keyStroke.getCharacter());
            inputLine += keyStroke.getCharacter();

            advanceCursor(cursorPosition);
        }
    }

    private void takeNap() {
        try {
            Thread.sleep(1);
        } catch (InterruptedException e) {
            isFinished = true;
        }
    }

    public void writeOutput() {
        try {
            for (int c = outputReader.read(); c != EOF; c = outputReader.read())
                processOutput((char) c);
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        terminal.setCursorVisible(true);
        terminal.flush();
        terminal.close();
    }

    private synchronized void processOutput(char c) {
        if (isEscape(c))
            parseControlSequence();
        else if (isEndOfSegment(c))
            printSegment();
        else
            outputSegment += c;
    }

    private synchronized void parseControlSequence() {}

    private synchronized boolean isEndOfSegment(char c) {
        return c == END_OF_SEGMENT;
    }

    private synchronized void printSegment() {
        terminal.setCursorVisible(false);

        for (char c : outputSegment.toCharArray())
            terminal.putCharacter(c);

        terminal.flush();
        terminal.setCursorVisible(true);
        outputSegment = "";
        origin = terminal.getCursorPosition();
    }

    public void finish() {
        isFinished = true;

        try {
            inputWriter.close();
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

}
